/*
 * Main.java
 * Version 1.0
 *
 * Created on March 26, 2008, 10:22 PM
 * Last modified on March 30, 2008.
 * McGill University
 */

package jmirutilities;

import java.io.*;
import ace.datatypes.*;
import jMusicMetaManager.RecordingMetaData;


/**
 * This utility performs various utility tasks for the jMIR package, generally
 * related to generating labelled ACE XML Classification Files, file parsing,
 * file type conversions and combining existing ARFF or ACE XML files in various
 * ways. It can only be run from the command line. The first command line
 * argument must be a flag with a hyphen preceding it, and must be one of the
 * valid choices specified in the help. Such a flag specifies the task to be
 * performed. Some flags permit additional command line options afterwards,
 * which do not require hyphens. Flags, associated functionality and options can
 * be displayed by running this with  the "-help" flag.
 *
 * @author Cory McKay
 */
public class Main
{
     /**
      * Runs the software from the command line. Run with the arguments "help"
      * or "-help" for more information
      *
      * @param args  The command line arguments.
      */
     public static void main(String[] args)
     {
          // Output an error if not flags are provided
          if (args.length == 0)
          {
               listOptions(true);
               System.exit(-1);
          }
          
          // Bring up a GUI for labelling instances based on file paths
          if(args[0].equals("-fileLabelInstances"))
          {
               InstanceLabeller labeller = new InstanceLabeller();
          }
          
          // Label instances based on a tab delimited text file
          else if(args[0].equals("-txtLabelInstances"))
          {
               // Extract information from the command line
               String input_file_path = null;
               if (args.length > 1) input_file_path = args[1];
               String output_file_path = null;
               if (args.length > 2) output_file_path = args[2];
               String[] field_types = null;
               
               // Extract data and save it
               tabDelimitedToACEXML(input_file_path, output_file_path);
          }
          
          // Merge extracted features
          else if(args[0].equals("-mergeFeatures"))
          {
               // Extract information from the command line
               String input_feature_vector_file_path_1 = null;
               if (args.length > 1) input_feature_vector_file_path_1 = args[1];
               String input_feature_definitions_file_path_1 = null;
               if (args.length > 2) input_feature_definitions_file_path_1 = args[2];
               String input_feature_vector_file_path_2 = null;
               if (args.length > 3) input_feature_vector_file_path_2 = args[3];
               String input_feature_definitions_file_path_2 = null;
               if (args.length > 4) input_feature_definitions_file_path_2 = args[4];
               String matching_key_file_path = null;
               if (args.length > 5) matching_key_file_path = args[5];
               String output_feature_vector_file_path = null;
               if (args.length > 6) output_feature_vector_file_path = args[6];
               String output_feature_definitions_file_path = null;
               if (args.length > 7) output_feature_definitions_file_path = args[7];
               
               // Extract data and save it
               mergeFeatures(input_feature_vector_file_path_1,
                    input_feature_definitions_file_path_1,
                    input_feature_vector_file_path_2,
                    input_feature_definitions_file_path_2,
                    matching_key_file_path,
                    output_feature_vector_file_path,
                    output_feature_definitions_file_path );
          }
          
          // Extract fields from an iTunes file and output them to a text file
          else if (args[0].equals("-matchItunesFields"))
          {
               // Extract information from the command line
               String input_file_path = null;
               if (args.length > 1) input_file_path = args[1];
               String output_file_path = null;
               if (args.length > 2) output_file_path = args[2];
               String[] field_types = null;
               if (args.length > 3)
               {
                    field_types = new String[args.length - 3];
                    for (int i = 3; i < args.length; i++)
                         field_types[i-3] = args[i];
               }
               
               // Extract data and save it
               matchItunesFields(input_file_path, output_file_path, field_types);
          }
          
          // List the available command line arguments
          else if (args[0].equals("-help") || args[0].equals("help"))
          {
               listOptions(false);
          }
          
          // Indicate to the user that valid command line options were not provided
          else
          {
               listOptions(true);
          }
     }
     
     
     /* PRIVATE METHODS *******************************************************/
     
     
     /**
      * Generate a model classifications ACE XML Classifications File based on
      * the specified tab delimited text input file. Each instance corresponds
      * to a line in the text file. The first entry on the line is the name of
      * the instance and the second (or more) are the name(s) of its class(es).
      *
      * <p>Multiple labels may be assigned to an instance, but this method
      * does not allow hierarchical labelling (although there is nothing in the
      * underlying ACE classes that prevents it).
      *
      * <p>Errors are output to the command line.
      *
      * @param   input_file_path   The path of the text file to parse.
      * @param   output_file_path  The path of the ACE XML file to generate.
      */
     private static void tabDelimitedToACEXML(String input_file_path,
          String output_file_path)
     {
          try
          {
               // Do error checking and open the input file
               File input_file = null;
               if (input_file_path == null) throw new Exception("No input file specified.");
               if (output_file_path == null) throw new Exception("No output file specified.");
               
               // Open the input and output files file
               try
               {
                    input_file = new File(input_file_path);
               }
               catch (Exception e)
               {throw new Exception("Unable to open input file at the specified path: " + input_file_path);}
               
               // Parse the input text file
               String[] lines_of_file = mckay.utilities.staticlibraries.FileMethods.parseTextFileLines(input_file);
               String[][] parsed_data = new String[lines_of_file.length][];
               for (int line = 0; line < lines_of_file.length; line++)
               {
                    try
                    {
                         parsed_data[line] = mckay.utilities.staticlibraries.StringMethods.breakIntoTokens(lines_of_file[line], "\t");
                    }
                    catch (Exception e)
                    {
                         throw new Exception("Unable to parse the specified file: " + input_file_path + "\nProblem occured at line " + (line+1) + ": " + lines_of_file[line]);
                    }
               }
               
               // Prepare the array of model classifications
               SegmentedClassification[] combined_classifications = new SegmentedClassification[parsed_data.length];
               
               // Construct model classifications
               try
               {
                    // Add information instance by instance
                    for (int instance = 0; instance < combined_classifications.length; instance++)
                    {
                         // Prepare the instance
                         combined_classifications[instance] = new SegmentedClassification();
                         
                         // Add the instance label
                         combined_classifications[instance].identifier = parsed_data[instance][0];
                         
                         // Add the class name(s)
                         if (parsed_data[instance].length > 1)
                         {
                              String[] classes = new String[parsed_data[instance].length - 1];
                              for (int class_index = 0; class_index < classes.length; class_index++)
                                   classes[class_index] = parsed_data[instance][class_index + 1];
                              combined_classifications[instance].classifications = classes;
                         }
                    }
               }
               catch (Exception e)
               {
                    throw new Exception("Unable to generate model classification from the input file: " + input_file_path);
               }
               
               // Save the model classifications
               try
               {
                    File output_file = new File(output_file_path);
                    SegmentedClassification.saveClassifications(combined_classifications, output_file, "");
               }
               catch (Exception e)
               {
                    throw new Exception("Unable to write to the specified output path: " + output_file_path);
               }
          }
          
          // Output any errors to the command line
          catch (Exception e)
          {
               System.out.println("ERROR occured.");
               System.out.println(e.getMessage() + "\n");
               // e.printStackTrace();
          }
     }
     
     
     /**
      * Merges the extracted feature contained in two different ACE XML Feature
      * Vector files and their corresponding Feature Definition Files. The two
      * input Feature Vector files should refer to different feature extracted
      * for the same instances. This could be useful, for example, for combining
      * features extracted from jAudio with features extracted with jWebMiner
      * for the same songs.
      *
      * <p>NOTE THAT THIS HAS NOT BEEN DESIGNED TO WORK WITH DATASETS THAT
      * HAVE SUBSETS YET (e.g. features separately extracted separately for
      * windows). Such subsets are currently just ignored.
      *
      * @param input_feature_vector_file_path_1   The first input feature vector
      *                                           file path.
      * @param input_feature_defs_file_path_1     The feature definitions file
      *                                           for this first feature vector
      *                                           file.
      * @param input_feature_vector_file_path_2   The second input feature
      *                                           vector file
      * @param input_feature_defs_file_path_2     The feature definitions file
      *                                           for this second feature vector
      *                                           file.
      * @param matching_key_file_path             A tab delimited text file,
      *                                           such as one that might be
      *                                           output by the
      *                                           matchItunesFields method,
      *                                           where each line corresponds to
      *                                           an instance and each line has
      *                                           three pieces of information
      *                                           separated by tabs, namely the
      *                                           instance identifier for the
      *                                           fist feature vector file, the
      *                                           corresponding instance
      *                                           identifier for the second
      *                                           feature vector file, and the
      *                                           instance identifier to use for
      *                                           the output file.
      * @param output_feature_vector_file_path    The file path to save the
      *                                           combined feature vector file
      *                                           to.
      * @param output_feature_defs_file_path      The file path to save the
      *                                           combined feature definitions
      *                                           file to
      */
     private static void mergeFeatures(String input_feature_vector_file_path_1,
          String input_feature_defs_file_path_1,
          String input_feature_vector_file_path_2,
          String input_feature_defs_file_path_2,
          String matching_key_file_path,
          String output_feature_vector_file_path,
          String output_feature_defs_file_path )
     {
          try
          {
               // Make sure that all parameters are specified
               if (input_feature_vector_file_path_1 == null) throw new Exception("Input feature vector file 1 file path not specified. Seven options must be present for this operation.");
               if (input_feature_defs_file_path_1 == null) throw new Exception("Input feature definitions file 1 file path not specified. Seven options must be present for this operation.");
               if (input_feature_vector_file_path_2 == null) throw new Exception("Input feature vector file 2 file path not specified. Seven options must be present for this operation.");
               if (input_feature_defs_file_path_2 == null) throw new Exception("Input feature definitions file 2 file path not specified. Seven options must be present for this operation.");
               if (matching_key_file_path == null) throw new Exception("Matching key file path not specified. Seven options must be present for this operation.");
               if (output_feature_vector_file_path == null) throw new Exception("Output feature vector file path not specified. Seven options must be present for this operation.");
               if (output_feature_defs_file_path == null) throw new Exception("Output feature definitions file path not specified. Seven options must be present for this operation.");
               
               // Parse the matching key file
               File matching_key_file  = null;
               try
               {matching_key_file = new File(matching_key_file_path);}
               catch (Exception e)
               {throw new Exception("Unable to access an input file at the specified path: " + matching_key_file_path);}
               String[] matching_key_lines_of_file = mckay.utilities.staticlibraries.FileMethods.parseTextFileLines(matching_key_file);
               String[][] matching_key_data = new String[matching_key_lines_of_file.length][];
               for (int line = 0; line < matching_key_lines_of_file.length; line++)
               {
                    try
                    {matching_key_data[line] = mckay.utilities.staticlibraries.StringMethods.breakIntoTokens(matching_key_lines_of_file[line], "\t");}
                    catch (Exception e)
                    {throw new Exception("Unable to parse the matching key file: " + matching_key_file_path + "\nProblem occured at line " + (line+1) + ": " + matching_key_lines_of_file[line]);}
                    if (matching_key_data[line].length != 3)
                         throw new Exception("Unable to parse the matching key file: " + matching_key_file_path + "\nThere must be 3 pieces of data on each line, separated by a tab, and this file has a line with " + matching_key_data[line].length + " pieces of data.\nProblem occured at line " + (line+1) + ": " + matching_key_lines_of_file[line]);
               }
               
               // Parse the two feature definition files
               FeatureDefinition[] feature_definitions_1 = null;
               try
               {feature_definitions_1 = ace.datatypes.FeatureDefinition.parseFeatureDefinitionsFile(input_feature_defs_file_path_1);}
               catch (Exception e)
               {throw new Exception("Unable to parse the first specified feature definitions file at the specified path: " + input_feature_defs_file_path_1 + "\n\nDETAILS:\n" + e.getMessage());}
               FeatureDefinition[] feature_definitions_2 = null;
               try
               {feature_definitions_2 = ace.datatypes.FeatureDefinition.parseFeatureDefinitionsFile(input_feature_defs_file_path_2);}
               catch (Exception e)
               {throw new Exception("Unable to parse second specified feature definitions file at the specified path: " + input_feature_defs_file_path_2 + "\n\nDETAILS:\n" + e.getMessage());}
               
               // Combine the two feature definition files
               FeatureDefinition[][] temp_definition_array = {feature_definitions_1, feature_definitions_2};
               FeatureDefinition[] feature_definitions_combined = ace.datatypes.FeatureDefinition.getMergedFeatureDefinitions(temp_definition_array);
               
               // Parse the two feature vector files
               DataSet[][] feature_vectors_array = new DataSet[2][];
               try
               {feature_vectors_array[0] = ace.datatypes.DataSet.parseDataSetFile(input_feature_vector_file_path_1, feature_definitions_1);}
               catch (Exception e)
               {throw new Exception("Unable to parse first specified feature values file at the specified path: " + input_feature_vector_file_path_1 + "\n\nDETAILS:\n" + e.getMessage());}
               try
               {feature_vectors_array[1] = ace.datatypes.DataSet.parseDataSetFile(input_feature_vector_file_path_2, feature_definitions_2);}
               catch (Exception e)
               {throw new Exception("Unable to parse second specified feature values file at the specified path: " + input_feature_vector_file_path_2 + "\n\nDETAILS:\n" + e.getMessage());}
               
               // Combine the feature vectos
               DataSet[] feature_vectors_combined = ace.datatypes.DataSet.getMergedFeatureTypes(feature_vectors_array, feature_definitions_combined, matching_key_data);
               
               // Save the combined feature definitions
               ace.datatypes.FeatureDefinition.saveFeatureDefinitions(feature_definitions_combined, new File(output_feature_defs_file_path), "Combined feature definitions of " + input_feature_defs_file_path_1 + " and " + input_feature_defs_file_path_2 + ".");
               
               // Save the combined feature vectors
               ace.datatypes.DataSet.saveDataSets(feature_vectors_combined, feature_definitions_combined, new File(output_feature_vector_file_path), "Combined feature vectors of " + input_feature_vector_file_path_1 + " and " + input_feature_vector_file_path_2 + ".");
          }
          
          // Output any errors to the command line
          catch (Exception e)
          {
               System.out.println("ERROR occured.");
               System.out.println(e.getMessage() + "\n");
               // e.printStackTrace();
          }
     }
     
     
     /**
      * Parse an iTunes XML file and output a text file where each line
      * corresponds to a recording and the tab delimited entries for that line
      * correspond to the field entries for the specified fields for the iTunes
      * file.
      *
      * <p>Fields types may be "path", "title", "artist", "composer", "album"
      * or "genre".
      *
      * <p>Errors are output to the command line.
      *
      * @param   input_file_path     The path of the iTunes XML file to parse.
      * @param   output_file_path    The path of the text file to generate.
      * @param   field_types         An array of field type identifiers.
      */
     private static void matchItunesFields(String input_file_path,
          String output_file_path,
          String[] field_types)
     {
          // Set identifiers for iTunes fields
          final String FILE_PATH_ID = "path";
          final String TITLE_ID = "title";
          final String ARTIST_ID = "artist";
          final String COMPOSER_ID = "composer";
          final String ALBUM_ID = "album";
          final String GENRE_ID = "genre";
          
          try
          {
               // Do error checking and open the input file
               File input_file = null;
               if (input_file_path == null) throw new Exception("No input file specified.");
               if (output_file_path == null) throw new Exception("No output file specified.");
               if (field_types == null) throw new Exception("No field types to parse specified.");
               if (field_types.length > 6) throw new Exception("You have specified " + field_types.length + " field types. The maximum is 6.");
               
               // Parse the input file
               RecordingMetaData[] itunes_data = null;
               try
               {
                    input_file = new File(input_file_path);
                    itunes_data = RecordingMetaData.extractMetaDataFromiTunesXML(input_file, false);
               }
               catch (Exception e)
               {throw new Exception("Unable to open input file at the specified path: " + input_file_path);}
               
               // To hold the matched strings. The first dimension is the recording
               // and the second is the field.
               String[][] data_to_return = new String[itunes_data.length][field_types.length];
               
               // Extract the first field
               for (int recording = 0; recording < itunes_data.length; recording++)
               {
                    for (int field = 0; field < field_types.length; field++)
                    {
                         if (field_types[field].equals(FILE_PATH_ID))
                              data_to_return[recording][field] = itunes_data[recording].file_path;
                         else if (field_types[field].equals(TITLE_ID))
                              data_to_return[recording][field] = itunes_data[recording].title;
                         else if (field_types[field].equals(ARTIST_ID))
                              data_to_return[recording][field] = itunes_data[recording].artist;
                         else if (field_types[field].equals(COMPOSER_ID))
                              data_to_return[recording][field] = itunes_data[recording].composer;
                         else if (field_types[field].equals(ALBUM_ID))
                              data_to_return[recording][field] = itunes_data[recording].album;
                         else if (field_types[field].equals(GENRE_ID))
                         {
                              if (itunes_data[recording].genres != null)
                              {
                                   data_to_return[recording][field] = "";
                                   for (int j = 0; j < itunes_data[recording].genres.length; j++)
                                   {
                                        if (j != 0) data_to_return[recording][field] += " + ";
                                        data_to_return[recording][field] += itunes_data[recording].genres[j];
                                   }
                              }
                         }
                         else throw new Exception("Field type of " + field_types[field] + " chosen. This is not a valid field type.\nFields types may be \"path\", \"title\", \"artist\", \"composer\", \"album\" or \"genre\".");
                    }
               }
               
               // Write the extracted information
               try
               {
                    // Prepare to write
                    File output_file = mckay.utilities.staticlibraries.FileMethods.getNewFileForWriting(output_file_path, true);
                    DataOutputStream writer = mckay.utilities.staticlibraries.FileMethods.getDataOutputStream(output_file);
                    
                    // Write lines one by one
                    for (int recording = 0; recording < data_to_return.length; recording++)
                    {
                         for (int field = 0; field < field_types.length; field++)
                         {
                              // Insert the delimiting tab if appropriate
                              if (field != 0) writer.writeBytes("\t");
                              
                              // Write the contents of this field
                              if (data_to_return[recording][field] != null)
                                   writer.writeBytes(data_to_return[recording][field]);
                         }
                         
                         // End the line
                         if (recording != data_to_return.length - 1) writer.writeBytes("\n");
                    }
                    
                    // Finish writing
                    writer.close();
               }
               catch (Exception e)
               {
                    throw new Exception("Unable to write to the specified output path: " + output_file_path);
               }
          }
          
          // Output any errors to the command line
          catch (Exception e)
          {
               System.out.println("ERROR occured.");
               System.out.println(e.getMessage() + "\n");
               // e.printStackTrace();
          }
     }
     
     
     /**
      * Prints the available choices of command line choices to standard out.
      *
      * @param command_line_error    Whether or not to print a message saying
      *                              that invalid command line parameters were
      *                              specified.
      */
     private static void listOptions(boolean command_line_error)
     {
          if (command_line_error) System.out.println("ERROR: Valid command line options were not provided.");
          
          System.out.println("\nThis utility performs various utility tasks for the jMIR package, generally related to generating labelled ACE XML Classification Files, file parsing, file type conversions and combining existing ARFF or ACE XML files in various ways. It can only be run from the command line. The first command line argument must be a flag with a hyphen preceding it, and must be one of the choices listed below. Such a flag specifies the task to be performed. Some flags permit additional arguments afterwards, which do not require hyphens.");
          System.out.println("\nLABELLING OPTIONS:");
          System.out.println("-fileLabelInstances: Run a GUI for generating an ACE XML Classifications File to contain model labels where instances are identified based on file paths batch selected by the user, and assigned class names are batch entered by the user.");
          System.out.println("-txtLabelInstances: Generate a model classifications ACE XML Classifications File based on a tab delimited text file. Each instance corresponds to a line in the text file. The first entry on the line is the name of the instance and the second (or more) are the name(s) of its class(es). The first command line option must specify the input file path of the tab delimited text file and the second must specify the output file path of the ACE XML file to be saved.");
          System.out.println("\nMERGE OPTIONS:");
          System.out.println("-mergeFeatures: Merge the extracted feature contained in two different ACE XML Feature Vector files and their corresponding Feature Definition Files. The two input Feature Vector files should refer to different feature extracted for the same instances. This could be useful, for example, for combining features extracted from jAudio with features extracted with jWebMiner for the same songs. There should be seven options provided at the command line, each referring to a file path: 1) The first input feature vector file. 2) The feature definitions file for this first feature vector file. 3) The second input feature vector file. 4) The feature definitions file for this second feature vector file. 5) A tab delimited text file, such as one that might be output by the -matchItunesFields flag, where each line corresponds to an instance and each line has three pieces of information separated by tabs, namely the instance identifier for the fist feature vector file, the corresponding instance identifier for the second feature vector file, and the instance identifier to use for the output file. 6) The file path to save the combined feature vector file to. 7) The file path to save the combined feature definitions file to. Note that this cannot yet deal with datasets that have subsets (e.g. features separately extracted separately for windows), and such subsets are currently just ignored.");
          System.out.println("\nFILE PARSING OPTIONS:");
          System.out.println("-matchItunesFields: Parse an iTunes XML file and output a text file where each line corresponds to a recording and the tab delimited entries for that line correspond to the field entries for the specified fields for the iTunes file. Fields types may be \"path\", \"title\", \"artist\", \"composer\", \"album\" or \"genre\". Three to eight command line options must follow this flag. The first must be the input file path of the iTunes XML file, the second must be the output file path of the text file to be saved, and the remainder specify the field type(s) and in what order they are to be listed in the text file.");
          System.out.println("\nHELP OPTIONS:");
          System.out.println("-help: List the command line interface options.");
          System.out.println("\n");
     }
}